@@ -290,6 +290,7 @@ struct resp_pkt {
 #define	INFO_FLAG_AUTHENABLE	0x20
 #define	INFO_FLAG_SEL_CANDIDATE	0x40
 #define	INFO_FLAG_SHORTLIST	0x80
+#define	INFO_FLAG_IBURST	0x100
 
 /*
  * Flags in the system information returns
@@ -517,7 +518,7 @@ struct info_mem_stats {
 	u_int32 findpeer_calls;
 	u_int32 allocations;
 	u_int32 demobilizations;
-	u_char hashcount[HASH_SIZE];
+	u_char hashcount[NTP_HASH_SIZE];
 };
 
 
