@@ -21,6 +21,12 @@
 #include "ntp_select.h"
 #include "ntp_stdlib.h"
 
+/* Don't include ISC's version of IPv6 variables and structures */
+#define ISC_IPV6_H 1
+#include "isc/net.h"
+#include "isc/result.h"
+#include "isc/sockaddr.h"
+
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
@@ -32,7 +38,6 @@
 # include <poll.h>
 #endif
 #ifndef SYS_WINNT
-# include <netdb.h>
 # ifdef HAVE_SYS_SIGNAL_H
 #  include <sys/signal.h>
 # else
@@ -48,10 +53,6 @@
 
 #include <arpa/inet.h>
 
-#ifdef __QNXNTO__
-# include "adjtime.h"
-#endif
-
 #ifdef SYS_VXWORKS
 # include "ioLib.h"
 # include "sockLib.h"
@@ -59,6 +60,12 @@
 
 /* select wants a zero structure ... */
 struct timeval timeout = {0,0};
+#elif defined(SYS_WINNT)
+/*
+ * Windows does not abort a select select call if SIGALRM goes off
+ * so a 200 ms timeout is needed
+ */
+struct timeval timeout = {0,1000000/TIMER_HZ};
 #else
 struct timeval timeout = {60,0};
 #endif
@@ -71,11 +78,12 @@ struct timeval timeout = {60,0};
 
 #ifdef SYS_WINNT
 #define EPROTONOSUPPORT WSAEPROTONOSUPPORT
-#define EAFNOSUPPORT    WSAEAFNOSUPPORT
-#define EPFNOSUPPORT    WSAEPFNOSUPPORT
+#define EAFNOSUPPORT	WSAEAFNOSUPPORT
+#define EPFNOSUPPORT	WSAEPFNOSUPPORT
 #define TARGET_RESOLUTION 1  /* Try for 1-millisecond accuracy
 				on Windows NT timers. */
 #pragma comment(lib, "winmm")
+isc_boolean_t ntp_port_inuse(int af, u_short port);
 #endif /* SYS_WINNT */
 
 /*
@@ -103,6 +111,11 @@ static timer_t ntpdate_timerid;
 #define NTP_MAXLIST 5	/* maximum select list size */
 #define PEER_SHIFT	8	/* 8 suitable for crystal time base */
 
+/*
+ * for get_systime()
+ */
+s_char	sys_precision;		/* local clock precision (log2 s) */
+
 /*
  * Debugging flag
  */
@@ -120,7 +133,7 @@ int fd_family[MAX_AF];	/* to remember the socket family */
 struct pollfd fdmask[MAX_AF];
 #else
 fd_set fdmask;
-int maxfd;
+SOCKET maxfd;
 #endif
 int polltest = 0;
 
@@ -319,12 +332,12 @@ ntpdatemain (
 	)
 {
 	int was_alarmed;
-	struct recvbuf *rbuflist;
+	int tot_recvbufs;
 	struct recvbuf *rbuf;
 	l_fp tmp;
 	int errflg;
 	int c;
-        int nfound;
+	int nfound;
 
 #ifdef HAVE_NETINFO
 	ni_namelist *netinfoservers;
@@ -350,6 +363,12 @@ ntpdatemain (
 	clear_globals();
 #endif
 
+
+	/* Check to see if we have IPv6. Otherwise force the -4 flag */
+	if (isc_net_probeipv6() != ISC_R_SUCCESS) {
+		ai_fam_templ = AF_INET;
+	}
+
 	errflg = 0;
 	progname = argv[0];
 	syslogit = 0;
@@ -463,7 +482,12 @@ ntpdatemain (
 	if (debug || simple_query) {
 #ifdef HAVE_SETVBUF
 		static char buf[BUFSIZ];
+#ifdef SYS_WINNT
+		/* Win32 does not implement line bufferibg */
+		setvbuf(stdout, NULL, _IONBF, BUFSIZ);
+#else
 		setvbuf(stdout, buf, _IOLBF, BUFSIZ);
+#endif	/* SYS_WINNT */
 #else
 		setlinebuf(stdout);
 #endif
@@ -569,45 +593,41 @@ ntpdatemain (
 
 	initializing = 0;
 	was_alarmed = 0;
-	rbuflist = (struct recvbuf *)0;
 
 	while (complete_servers < sys_numservers) {
 #ifdef HAVE_POLL_H
-                struct pollfd* rdfdes;
-                rdfdes = fdmask;
+		struct pollfd* rdfdes;
+		rdfdes = fdmask;
 #else
 		fd_set rdfdes;
-                rdfdes = fdmask;
+		rdfdes = fdmask;
 #endif
 
 		if (alarm_flag) {		/* alarmed? */
 			was_alarmed = 1;
 			alarm_flag = 0;
 		}
-		rbuflist = getrecvbufs();	/* get received buffers */
+		tot_recvbufs = full_recvbuffs();	/* get received buffers */
 
-		if (!was_alarmed && rbuflist == (struct recvbuf *)0) {
+		if (!was_alarmed && tot_recvbufs == 0) {
 			/*
 			 * Nothing to do.	 Wait for something.
 			 */
 #ifdef HAVE_POLL_H
-                        nfound = poll(rdfdes, (unsigned int)nbsock, timeout.tv_sec * 1000);
+			nfound = poll(rdfdes, (unsigned int)nbsock, timeout.tv_sec * 1000);
 
 #else
-                        nfound = select(maxfd, &rdfdes, (fd_set *)0,
+			nfound = select(maxfd, &rdfdes, (fd_set *)0,
 					(fd_set *)0, &timeout);
 #endif
 			if (nfound > 0)
 				input_handler();
-			else if (
-#ifndef SYS_WINNT
-				nfound == -1
-#else
-				nfound == SOCKET_ERROR
-#endif /* SYS_WINNT */
-				) {
+			else if (nfound == SOCKET_ERROR)
+			{
 #ifndef SYS_WINNT
 				if (errno != EINTR)
+#else
+				if (WSAGetLastError() != WSAEINTR)
 #endif
 					netsyslog(LOG_ERR,
 #ifdef HAVE_POLL_H
@@ -616,7 +636,7 @@ ntpdatemain (
 						"select() error: %m"
 #endif
 						);
-			} else {
+			} else if (errno != 0) {
 #ifndef SYS_VXWORKS
 				netsyslog(LOG_DEBUG,
 #ifdef HAVE_POLL_H
@@ -631,18 +651,19 @@ ntpdatemain (
 				was_alarmed = 1;
 				alarm_flag = 0;
 			}
-			rbuflist = getrecvbufs();	/* get received buffers */
+			tot_recvbufs = full_recvbuffs();	/* get received buffers */
 		}
 
 		/*
 		 * Out here, signals are unblocked.  Call receive
 		 * procedure for each incoming packet.
 		 */
-		while (rbuflist != (struct recvbuf *)0) {
-			rbuf = rbuflist;
-			rbuflist = rbuf->next;
+		rbuf = get_full_recv_buffer();
+		while (rbuf != NULL)
+		{
 			receive(rbuf);
 			freerecvbuf(rbuf);
+			rbuf = get_full_recv_buffer();
 		}
 
 		/*
@@ -676,8 +697,8 @@ ntpdatemain (
 
 /*
  * transmit - transmit a packet to the given server, or mark it completed.
- *		  This is called by the timeout routine and by the receive
- *		  procedure.
+ *		This is called by the timeout routine and by the receive
+ *		procedure.
  */
 static void
 transmit(
@@ -710,7 +731,7 @@ transmit(
 	}
 
 	/*
-	 * If we're here, send another message to the server.    Fill in
+	 * If we're here, send another message to the server.  Fill in
 	 * the packet and let 'er rip.
 	 */
 	xpkt.li_vn_mode = PKT_LI_VN_MODE(LEAP_NOTINSYNC,
@@ -779,7 +800,7 @@ receive(
 	int is_authentic;
 
 	if (debug)
-                printf("receive(%s)\n", stoa(&rbufp->recv_srcadr));
+		printf("receive(%s)\n", stoa(&rbufp->recv_srcadr));
 	/*
 	 * Check to see if the packet basically looks like something
 	 * intended for us.
@@ -1060,29 +1081,29 @@ clock_select(void)
 		if (server->delay == 0) {
 			if (debug)
 				printf("%s: Server dropped: no data\n", ntoa(&server->srcadr));
-			continue;   /* no data */
+			continue;	/* no data */
 		}
 		if (server->stratum > NTP_INFIN) {
 			if (debug)
 				printf("%s: Server dropped: strata too high\n", ntoa(&server->srcadr));
-			continue;   /* stratum no good */
+			continue;	/* stratum no good */
 		}
 		if (server->delay > NTP_MAXWGT) {
 			if (debug)
 				printf("%s: Server dropped: server too far away\n", 
-				       ntoa(&server->srcadr));
-			continue;   /* too far away */
+					ntoa(&server->srcadr));
+			continue;	/* too far away */
 		}
 		if (server->leap == LEAP_NOTINSYNC) {
 			if (debug)
 				printf("%s: Server dropped: Leap not in sync\n", ntoa(&server->srcadr));
-			continue;   /* he's in trouble */
+			continue;	/* he's in trouble */
 		}
 		if (!L_ISHIS(&server->org, &server->reftime)) {
 			if (debug)
 				printf("%s: Server dropped: server is very broken\n", 
 				       ntoa(&server->srcadr));
-			continue;   /* very broken host */
+			continue;	/* very broken host */
 		}
 		if ((server->org.l_ui - server->reftime.l_ui)
 		    >= NTP_MAXAGE) {
@@ -1239,7 +1260,7 @@ clock_select(void)
 	}
 
 	/*
-	 * That's it.    Return our server.
+	 * That's it.  Return our server.
 	 */
 	return sys_server;
 }
@@ -1313,10 +1334,34 @@ clock_adjust(void)
 }
 
 
+/*
+ * is_unreachable - check to see if we have a route to given destination
+ *		    (non-blocking).
+ */
+static int
+is_reachable (struct sockaddr_storage *dst)
+{
+	SOCKET sockfd;
+
+	sockfd = socket(dst->ss_family, SOCK_DGRAM, 0);
+	if (sockfd == -1) {
+		return 0;
+	}
+
+	if(connect(sockfd, (struct sockaddr *)dst, SOCKLEN(dst))) {
+		closesocket(sockfd);
+		return 0;
+	}
+	closesocket(sockfd);
+	return 1;
+}
+
+
+
 /* XXX ELIMINATE: merge BIG slew into adj_systime in lib/systime.c */
 /*
  * addserver - determine a server's address and allocate a new structure
- *		   for it.
+ *		for it.
  */
 static void
 addserver(
@@ -1324,49 +1369,60 @@ addserver(
 	)
 {
 	register struct server *server;
-        /* Address infos structure to store result of getaddrinfo */
-        struct addrinfo *addrResult;
-        /* Address infos structure to store hints for getaddrinfo */
-        struct addrinfo hints;
-        /* Error variable for getaddrinfo */
-        int error;
-        /* Service name */
-        char service[5];
+	/* Address infos structure to store result of getaddrinfo */
+	struct addrinfo *addrResult, *ptr;
+	/* Address infos structure to store hints for getaddrinfo */
+	struct addrinfo hints;
+	/* Error variable for getaddrinfo */
+	int error;
+	/* Service name */
+	char service[5];
 	strcpy(service, "ntp");
 
-        /* Get host address. Looking for UDP datagram connection. */
-        memset(&hints, 0, sizeof(hints));
-        hints.ai_family = ai_fam_templ;
-        hints.ai_socktype = SOCK_DGRAM;
+	/* Get host address. Looking for UDP datagram connection. */
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_family = ai_fam_templ;
+	hints.ai_socktype = SOCK_DGRAM;
 
-        printf("Looking for host %s and service %s\n", serv, service);
+#ifdef DEBUG
+	if (debug)
+		printf("Looking for host %s and service %s\n", serv, service);
+#endif
 
-        error = getaddrinfo(serv, service, &hints, &addrResult);
-        if (error != 0) {
-                fprintf(stderr, "Error : %s\n", gai_strerror(error));
+	error = getaddrinfo(serv, service, &hints, &addrResult);
+	if (error != 0) {
+		fprintf(stderr, "Error : %s\n", gai_strerror(error));
 		msyslog(LOG_ERR, "can't find host %s\n", serv);
 		return;
 	}
-        else {
-                fprintf(stderr, "host found : %s\n", stohost((struct sockaddr_storage*)addrResult->ai_addr));
-        }
-
-	server = (struct server *)emalloc(sizeof(struct server));
-	memset((char *)server, 0, sizeof(struct server));
-
-        /* For now we only get the first returned server of the addrinfo list */
-        memset(&(server->srcadr), 0, sizeof(struct sockaddr_storage));
-        memcpy(&(server->srcadr), addrResult->ai_addr, addrResult->ai_addrlen);
-	server->event_time = ++sys_numservers;
-	if (sys_servers == NULL)
-		sys_servers = server;
-	else {
-		struct server *sp;
-
-		for (sp = sys_servers; sp->next_server != NULL;
-		     sp = sp->next_server) ;
-		sp->next_server = server;
+#ifdef DEBUG
+	else if (debug) {
+		fprintf(stderr, "host found : %s\n", stohost((struct sockaddr_storage*)addrResult->ai_addr));
 	}
+#endif
+
+	/* We must get all returned server in case the first one fails */
+	for (ptr = addrResult; ptr != NULL; ptr = ptr->ai_next) {
+		if (is_reachable ((struct sockaddr_storage *)ptr->ai_addr)) {
+			server = (struct server *)emalloc(sizeof(struct server));
+			memset((char *)server, 0, sizeof(struct server));
+
+			memset(&(server->srcadr), 0, sizeof(struct sockaddr_storage));
+			memcpy(&(server->srcadr), ptr->ai_addr, ptr->ai_addrlen);
+			server->event_time = ++sys_numservers;
+			if (sys_servers == NULL)
+				sys_servers = server;
+			else {
+				struct server *sp;
+
+				for (sp = sys_servers; sp->next_server != NULL;
+				     sp = sp->next_server) ;
+				sp->next_server = server;
+			}
+		}
+	}
+
+	freeaddrinfo(addrResult);
 }
 
 
@@ -1381,6 +1437,16 @@ findserver(
 {
 	struct server *server;
 	struct server *mc_server;
+	isc_sockaddr_t laddr;
+	isc_sockaddr_t saddr;
+
+	if(addr->ss_family == AF_INET) {
+		isc_sockaddr_fromin( &laddr, &((struct sockaddr_in*)addr)->sin_addr, 0);
+	}
+	else {
+		isc_sockaddr_fromin6(&laddr, &((struct sockaddr_in6*)addr)->sin6_addr, 0);
+	}
+
 
 	mc_server = NULL;
 	if (htons(((struct sockaddr_in*)addr)->sin_port) != NTP_PORT)
@@ -1389,21 +1455,19 @@ findserver(
 	for (server = sys_servers; server != NULL; 
 	     server = server->next_server) {
 		
-                if (memcmp(addr, &server->srcadr, SOCKLEN(addr))==0)
+		if(server->srcadr.ss_family == AF_INET) {
+			isc_sockaddr_fromin(&saddr, &((struct sockaddr_in*)&server->srcadr)->sin_addr, 0);
+		}
+		else {
+			isc_sockaddr_fromin6(&saddr, &((struct sockaddr_in6*)&server->srcadr)->sin6_addr, 0);
+		}
+		if (isc_sockaddr_eqaddr(&laddr, &saddr) == ISC_TRUE)
 			return server;
-                /* Multicast compatibility to verify here... I'm not sure it's working */
-                if(addr->ss_family == AF_INET) {
-                        if (IN_MULTICAST(ntohl(((struct sockaddr_in*)addr)->sin_addr.s_addr)))
-                                mc_server = server;
-                }
-                else {
-#ifdef AF_INET6
-                        if (IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6*)(&server->srcadr))->sin6_addr))
-			mc_server = server;
-#else
-                        return 0;
-#endif
-                }
+
+		if(addr->ss_family == server->srcadr.ss_family) {
+			if (isc_sockaddr_ismulticast(&saddr) == ISC_TRUE)
+				mc_server = server;
+		}
 	}
 
 	if (mc_server != NULL) {	
@@ -1418,7 +1482,7 @@ findserver(
 		server = (struct server *)emalloc(sizeof(struct server));
 		memset((char *)server, 0, sizeof(struct server));
 
-                memcpy(&server->srcadr, &addr, sizeof(struct sockaddr_storage));
+		memcpy(&server->srcadr, &addr, sizeof(struct sockaddr_storage));
 
 		server->event_time = ++sys_numservers;
 
@@ -1559,7 +1623,7 @@ init_alarm(void)
 	}
 	/* get the LUID for system-time privilege. */
 	LookupPrivilegeValue(NULL, SE_SYSTEMTIME_NAME, &tkp.Privileges[0].Luid);
-	tkp.PrivilegeCount = 1;  /* one privilege to set */
+	tkp.PrivilegeCount = 1;		/* one privilege to set */
 	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
 	/* get set-time privilege for this process. */
 	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0,(PTOKEN_PRIVILEGES) NULL, 0);
@@ -1581,7 +1645,7 @@ init_alarm(void)
 	}
 	/* get the LUID for system-time privilege. */
 	LookupPrivilegeValue(NULL, SE_SYSTEMTIME_NAME, &tkp.Privileges[0].Luid);
-	tkp.PrivilegeCount = 1;  /* one privilege to set */
+	tkp.PrivilegeCount = 1;		/* one privilege to set */
 	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
 	/* get set-time privilege for this process. */
 	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0,(PTOKEN_PRIVILEGES) NULL, 0);
@@ -1607,11 +1671,11 @@ init_alarm(void)
 
 	/* start the timer event */
 	wTimerID = timeSetEvent(
-		(UINT) (1000/TIMER_HZ),    /* Delay */
-		wTimerRes,			 /* Resolution */
-		(LPTIMECALLBACK) alarming, /* Callback function */
-		(DWORD) dwUser, 	 /* User data */
-		TIME_PERIODIC); 	 /* Event type (periodic) */
+		(UINT) (1000/TIMER_HZ),		/* Delay */
+		wTimerRes,			/* Resolution */
+		(LPTIMECALLBACK) alarming,	/* Callback function */
+		(DWORD) dwUser,			/* User data */
+		TIME_PERIODIC);			/* Event type (periodic) */
 	if (wTimerID == 0) {
 		msyslog(LOG_ERR, "timeSetEvent failed: %m");
 		exit(1);
@@ -1642,10 +1706,10 @@ init_alarm(void)
 static void
 init_io(void)
 {
-        struct addrinfo *res, *ressave;
-        struct addrinfo hints;
+	struct addrinfo *res, *ressave;
+	struct addrinfo hints;
 	char service[5];
-        int optval = 1;
+	int optval = 1;
 
 	/*
 	 * Init buffer free list and stat counters
@@ -1658,53 +1722,68 @@ init_io(void)
 
 	strcpy(service, "ntp");
 
-        /*
-         * Init hints addrinfo structure
-         */
-        memset(&hints, 0, sizeof(hints));
-        hints.ai_flags = AI_PASSIVE;
-        hints.ai_socktype = SOCK_DGRAM;
+	/*
+	 * Init hints addrinfo structure
+	 */
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_family = ai_fam_templ;
+	hints.ai_flags = AI_PASSIVE;
+	hints.ai_socktype = SOCK_DGRAM;
 
-        if(getaddrinfo(NULL, service, &hints, &res) != 0) {
-               msyslog(LOG_ERR, "getaddrinfo() failed: %m");
-               exit(1);
-               /*NOTREACHED*/
-        }
+	if(getaddrinfo(NULL, service, &hints, &res) != 0) {
+	       msyslog(LOG_ERR, "getaddrinfo() failed: %m");
+	       exit(1);
+	       /*NOTREACHED*/
+	}
 
-        /* Remember the address of the addrinfo structure chain */
-        ressave = res;
+#ifdef SYS_WINNT
+	if (ntp_port_inuse(AF_INET, NTP_PORT) ){
+		netsyslog(LOG_ERR, "the NTP socket is in use, exiting: %m");
+		exit(1);
+	}
+#endif
 
-        /*
-         * For each structure returned, open and bind socket
-         */
-        for(nbsock = 0; (nbsock < MAX_AF) && res ; res = res->ai_next) {
+	/* Remember the address of the addrinfo structure chain */
+	ressave = res;
+
+	/*
+	 * For each structure returned, open and bind socket
+	 */
+	for(nbsock = 0; (nbsock < MAX_AF) && res ; res = res->ai_next) {
 	/* create a datagram (UDP) socket */
-           if ((fd[nbsock] = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0) {
+	   fd[nbsock] = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
+	   if (fd[nbsock] == SOCKET_ERROR) {
+#ifndef SYS_WINNT
 		if (errno == EPROTONOSUPPORT || errno == EAFNOSUPPORT ||
 		    errno == EPFNOSUPPORT)
+#else
+		int err = WSAGetLastError();
+		if (err == WSAEPROTONOSUPPORT || err == WSAEAFNOSUPPORT ||
+		    err == WSAEPFNOSUPPORT)
+#endif
 			continue;
 		netsyslog(LOG_ERR, "socket() failed: %m");
 		exit(1);
 		/*NOTREACHED*/
-	}
-           /* set socket to reuse address */
-           if (setsockopt(fd[nbsock], SOL_SOCKET, SO_REUSEADDR, (void*) &optval, sizeof(optval)) < 0) {
-   		   netsyslog(LOG_ERR, "setsockopt() SO_REUSEADDR failed: %m");
-   		   exit(1);
+	   }
+	   /* set socket to reuse address */
+	   if (setsockopt(fd[nbsock], SOL_SOCKET, SO_REUSEADDR, (void*) &optval, sizeof(optval)) < 0) {
+		   netsyslog(LOG_ERR, "setsockopt() SO_REUSEADDR failed: %m");
+		   exit(1);
 		   /*NOTREACHED*/
-    	   }
+	   }
 #ifdef IPV6_V6ONLY
-           /* Restricts AF_INET6 socket to IPv6 communications (see RFC 2553bis-03) */
-           if (res->ai_family == AF_INET6)
-                if (setsockopt(fd[nbsock], IPPROTO_IPV6, IPV6_V6ONLY, (void*) &optval, sizeof(optval)) < 0) {
-   		           netsyslog(LOG_ERR, "setsockopt() IPV6_V6ONLY failed: %m");
-   		           exit(1);
-		           /*NOTREACHED*/
-    	        }
+	   /* Restricts AF_INET6 socket to IPv6 communications (see RFC 2553bis-03) */
+	   if (res->ai_family == AF_INET6)
+		if (setsockopt(fd[nbsock], IPPROTO_IPV6, IPV6_V6ONLY, (void*) &optval, sizeof(optval)) < 0) {
+			   netsyslog(LOG_ERR, "setsockopt() IPV6_V6ONLY failed: %m");
+			   exit(1);
+			   /*NOTREACHED*/
+		}
 #endif
 
-           /* Remember the socket family in fd_family structure */
-           fd_family[nbsock] = res->ai_family;
+	   /* Remember the socket family in fd_family structure */
+	   fd_family[nbsock] = res->ai_family;
 
 	/*
 	 * bind the socket to the NTP port
@@ -1729,9 +1808,9 @@ init_io(void)
 	    fdmask[nbsock].events = POLLIN;
 #else
 	    FD_SET(fd[nbsock], &fdmask);
-            if ((SOCKET) maxfd < fd[nbsock]+1) {
-                maxfd = fd[nbsock]+1;
-            }
+	    if (maxfd < fd[nbsock]+1) {
+		maxfd = fd[nbsock]+1;
+	    }
 #endif
 
 	/*
@@ -1742,7 +1821,7 @@ init_io(void)
   {
 	int on = TRUE;
 
-	   if (ioctl(fd[nbsock],FIONBIO, &on) == ERROR) {
+	if (ioctl(fd[nbsock],FIONBIO, &on) == ERROR) {
 	  netsyslog(LOG_ERR, "ioctl(FIONBIO) fails: %m");
 	  exit(1);
 	}
@@ -1772,7 +1851,7 @@ init_io(void)
 		exit(1);
 	}
 #endif /* SYS_WINNT */
-   	nbsock++;
+	nbsock++;
     }
     freeaddrinfo(ressave);
 }
@@ -1787,40 +1866,39 @@ sendpkt(
 	int len
 	)
 {
-        int i;
+	int i;
 	int cc;
-        SOCKET sock = 0;
+	SOCKET sock = INVALID_SOCKET;
 
 #ifdef SYS_WINNT
 	DWORD err;
 #endif /* SYS_WINNT */
 
-        /* Find a local family compatible socket to send ntp packet to ntp server */
-        for(i = 0; (i < MAX_AF); i++) {
-                if(dest->ss_family == fd_family[i]) {
-                        sock = fd[i];
-                break;
-                }
-        }
+	/* Find a local family compatible socket to send ntp packet to ntp server */
+	for(i = 0; (i < MAX_AF); i++) {
+		if(dest->ss_family == fd_family[i]) {
+			sock = fd[i];
+		break;
+		}
+	}
 
-        if ( sock == 0 ) {
-                netsyslog(LOG_ERR, "cannot find family compatible socket to send ntp packet");
-                exit(1);
-                /*NOTREACHED*/
-        }
+	if ( sock == INVALID_SOCKET ) {
+		netsyslog(LOG_ERR, "cannot find family compatible socket to send ntp packet");
+		exit(1);
+		/*NOTREACHED*/
+	}
 
 	cc = sendto(sock, (char *)pkt, len, 0, (struct sockaddr *)dest,
 			SOCKLEN(dest));
 
+	if (cc == SOCKET_ERROR) {
 #ifndef SYS_WINNT
-	if (cc == -1) {
 		if (errno != EWOULDBLOCK && errno != ENOBUFS)
 #else
-	if (cc == SOCKET_ERROR) {
 		err = WSAGetLastError();
 		if (err != WSAEWOULDBLOCK && err != WSAENOBUFS)
 #endif /* SYS_WINNT */
-                        netsyslog(LOG_ERR, "sendto(%s): %m", stohost(dest));
+			netsyslog(LOG_ERR, "sendto(%s): %m", stohost(dest));
 	}
 }
 
@@ -1836,13 +1914,13 @@ input_handler(void)
 	struct timeval tvzero;
 	int fromlen;
 	l_fp ts;
-        int i;
+	SOCKET i;
 #ifdef HAVE_POLL_H
 	struct pollfd fds[MAX_AF];
 #else
 	fd_set fds;
 #endif
-        int fdc = 0;
+	int fdc = 0;
 
 	/*
 	 * Do a poll to see if we have data
@@ -1851,33 +1929,33 @@ input_handler(void)
 		tvzero.tv_sec = tvzero.tv_usec = 0;
 #ifdef HAVE_POLL_H
 		memcpy(fds, fdmask, sizeof(fdmask));
-                n = poll(fds, (unsigned int)nbsock, tvzero.tv_sec * 1000);
+		n = poll(fds, (unsigned int)nbsock, tvzero.tv_sec * 1000);
 
-                /*
-                 * Determine which socket received data
-                 */
+		/*
+		 * Determine which socket received data
+		 */
 
-                 for(i=0; i < nbsock; i++) {
-                        if(fds[i].revents & POLLIN) {
-                                fdc = fd[i];
-                                break;
-                        }
-                 }
+		for(i=0; i < nbsock; i++) {
+			if(fds[i].revents & POLLIN) {
+				fdc = fd[i];
+				break;
+			}
+		}
 
 #else
 		fds = fdmask;
-                n = select(maxfd, &fds, (fd_set *)0, (fd_set *)0, &tvzero);
+		n = select(maxfd, &fds, (fd_set *)0, (fd_set *)0, &tvzero);
 
-                /*
-                 * Determine which socket received data
-                 */
+		/*
+		 * Determine which socket received data
+		 */
 
-                for(i=0; i < maxfd; i++) {
-                        if(FD_ISSET(fd[i], &fds)) {
-                                 fdc = fd[i];
-                                 break;
-                        }
-                }
+		for(i=0; i < maxfd; i++) {
+			if(FD_ISSET(fd[i], &fds)) {
+				 fdc = fd[i];
+				 break;
+			}
+		}
 
 #endif
 
@@ -2052,7 +2130,7 @@ l_step_systime(
 	l_adj_systime(ts);
 	return 1;
 #endif
-#else /* SLEWALWAYS  */
+#else /* SLEWALWAYS */
 	if (debug)
 		return 1;
 	LFPTOD(ts, dtemp);
@@ -2218,3 +2296,34 @@ getnetinfoservers(void)
 	return(namelist);
 }
 #endif
+
+#ifdef SYS_WINNT
+isc_boolean_t ntp_port_inuse(int af, u_short port)
+{
+	/*
+	 * Check if NTP socket is already in use on this system
+	 * This is only for Windows Systems, as they tend not to fail on the real bind() below
+	 */
+	
+	SOCKET checksocket;
+	struct sockaddr_in checkservice;
+	checksocket = socket(af, SOCK_DGRAM, 0);
+	if (checksocket == INVALID_SOCKET) {
+		return (ISC_TRUE);
+	}
+
+	checkservice.sin_family = (short) AF_INET;
+	checkservice.sin_addr.s_addr = INADDR_LOOPBACK;
+	checkservice.sin_port = htons(port);
+
+	if (bind(checksocket, (struct sockaddr *)&checkservice,
+		sizeof(checkservice)) == SOCKET_ERROR) {
+		if ( WSAGetLastError() == WSAEADDRINUSE ){
+			closesocket(checksocket);
+			return (ISC_TRUE);
+		}
+	}
+	closesocket(checksocket);
+	return (ISC_FALSE);
+}
+#endif
